/**
 * description
 *
 * @author eson
 *2022年7月12日-13:36:24
 */
package com.yuandian.dataflow.statemachine;
 
import java.io.File;
import java.nio.ByteBuffer;
import java.util.Set;
import java.util.concurrent.Executor;
import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.function.Function;

import org.reflections.Reflections;

import com.alipay.remoting.NamedThreadFactory;
import com.alipay.remoting.exception.CodecException;
import com.alipay.remoting.serialization.SerializerManager;
import com.alipay.sofa.jraft.JRaftUtils;
import com.alipay.sofa.jraft.Node;
import com.alipay.sofa.jraft.RaftGroupService;
import com.alipay.sofa.jraft.Status;
import com.alipay.sofa.jraft.closure.ReadIndexClosure;
import com.alipay.sofa.jraft.conf.Configuration;
import com.alipay.sofa.jraft.entity.PeerId;
import com.alipay.sofa.jraft.entity.Task;
import com.alipay.sofa.jraft.error.RaftError;
import com.alipay.sofa.jraft.error.RemotingException;
import com.alipay.sofa.jraft.option.CliOptions;
import com.alipay.sofa.jraft.option.NodeOptions;
import com.alipay.sofa.jraft.rpc.InvokeCallback;
import com.alipay.sofa.jraft.rpc.RpcClient;
import com.alipay.sofa.jraft.rpc.RpcProcessor;
import com.alipay.sofa.jraft.rpc.impl.BoltRaftRpcFactory;
import com.alipay.sofa.jraft.util.BytesUtil;
import com.alipay.sofa.jraft.util.Endpoint;
import com.alipay.sofa.jraft.util.ThreadPoolUtil;
import com.yuandian.dataflow.statemachine.rpc.ResponseSM;
import com.yuandian.dataflow.statemachine.rpc.SyncConditionProcessor.RequestCondition;
import com.yuandian.dataflow.statemachine.rpc.SyncStateProcessor.RequestState;
import com.yuandian.dataflow.statemachine.rpc.annotations.ProcessorRaft;
import com.yuandian.dataflow.statemachine.state.State;
import com.yuandian.dataflow.statemachine.state.WorkerState;

import lombok.Getter;
import lombok.Setter;
import lombok.var;
import lombok.extern.slf4j.Slf4j;

/**
 * description
 *
 * @author eson
 *2022年7月12日-13:36:24
 */
@Slf4j
@var
public class StateServerFactory {

    private static StateServer ss;
 
    private static String myPeerStr;
    private static Configuration raftConf;    
 
    public static void startStateServer(String peerstr, Configuration conf) throws Exception {
        if(ss != null) {
            throw new Exception("重复初始化 InitStateServer");
        }
        ss = new StateServerFactory.StateServer(peerstr, conf);
    }

 

 


    public static PeerId getServerId() {
        return ss.getCluster().getServerId();
    }

    public static Node getNode() {
        return ss.getNode() ;
    }

    public static RpcClient getRpcClient() {
        return ss.getRpcClient();
    }

    public static RaftGroupService getCluster() {
        return ss.getCluster();
    }

    // 获取状态服务的对象
    public static StateServer getStateServer() {
        return ss;
    }

    @Getter
    @Setter
    public static class StateServer {

    RpcClient rpcClient;

    private Node node;
    private RaftGroupService cluster;
    private StateMachine fsm;

    private String groupId = "dataflow";
    private Executor readIndexExecutor = createReadIndexExecutor(); 

    public StateServer(String addr, Configuration conf) {
        // String[] peers = new String[]{"localhost:4440","localhost:4441","localhost:4442"};
        // String[] sprPeers = new String[]{"3440","3441","3442"};

        // var peeridstr = peers[Integer.parseInt(serverId)];
        // var sprPort = sprPeers[Integer.parseInt(args[0])];

        // String groupId = "jraft";

        //  conf = JRaftUtils.getConfiguration("localhost:4440,localhost:4441,localhost:4442");

        PeerId serverId = JRaftUtils.getPeerId(addr);
        int port = serverId.getPort();

        NodeOptions nodeOptions = new NodeOptions();
        
        nodeOptions.setElectionTimeoutMs(1000);
        nodeOptions.setSnapshotLogIndexMargin(3600);
        nodeOptions.setInitialConf(conf);

        File RaftDataFile = new File(String.format("./raftdata/%d", port) );
        log.info("{}",RaftDataFile.mkdirs());  

        nodeOptions.setLogUri( String.format("./raftdata/%d/logs", port) );
        nodeOptions.setRaftMetaUri(String.format("./raftdata/%d/rafts", port));
        nodeOptions.setSnapshotUri(String.format("./raftdata/%d/snaps", port));
        fsm = new StateMachine(); // 状态实例初始化
        nodeOptions.setFsm(fsm);

        cluster = new RaftGroupService(groupId, serverId, nodeOptions);
        
        
         
        Set<Class<?>>  scans = new Reflections("com.yuandian.dataflow").getTypesAnnotatedWith(ProcessorRaft.class);
        scans.forEach((pRaftClass)->{
            try {
                cluster.getRpcServer().registerProcessor((RpcProcessor<?>) pRaftClass.newInstance());
            } catch (InstantiationException e) {
                e.printStackTrace();
            } catch (IllegalAccessException e) {
                e.printStackTrace();
            }
        });
        node = cluster.start();

        rpcClient = new BoltRaftRpcFactory().createRpcClient();
        rpcClient.init(new CliOptions());
    }

   

    public boolean isLeader() {
        return this.fsm.isLeader();
    }

 
    public void useFsmStateAsync(Function<State, Void> dofunc) { 

        SyncClosure<State> closure = new SyncClosure<State>() {
            @Override
            public void run(Status status) {
                if(status.isOk()) {
                    dofunc.apply(this.getValue());
                }
            }
         };
         
        getNode().readIndex(BytesUtil.EMPTY_BYTES, new ReadIndexClosure() {
            @Override
            public void run(Status status, long index, byte[] reqCtx) {

                getFsm().useState((fsmState)->{ 
                    if(status.isOk()){
                        closure.setValue(fsmState);
                        closure.success(fsmState);
                        closure.run(Status.OK()); 
                        return null;
                    }

                    readIndexExecutor.execute(() -> {
                        if(isLeader()){
                            log.info("Fail to get value with 'ReadIndex': {}, try to applying to the state machine.", getNode());
                            applyState(fsmState, closure);
                        }else {
                            handlerNotLeaderError(closure);
                        }
                    });
                    return null;
                });
            }
        });
 
    return ;
    }

    /**
     * 同步更新 WorkerState
     * @param dofunc
     * @throws InterruptedException
     * @throws RemotingException
     */
    public void updateFsmWorkerState(WorkerState ws) throws InterruptedException, RemotingException {

        // leader就直接提交
        if(isLeader()) {
            var closure = new SyncClosure<State>() {
                @Override
                public void run(Status status) {
                    // log.debug("leader {}", status);
                    this.synclock.notify();
                }
            };
            StateServerFactory.getStateServer().applyWorkerState(ws, closure); 
            closure.synclock.wait(5000);
            return;
        }

        try {
            // 非leader就 rpc请求
            var ss = StateServerFactory.getStateServer();
            var request = new RequestCondition();
            request.setWorkerState(ws);
            log.debug("my: {} leader id {}",ss.getCluster().getServerId(), ss.getNode().getLeaderId());
            ResponseSM resp;
            resp = (ResponseSM)StateServerFactory.getRpcClient().invokeSync(StateServerFactory.getNode().getLeaderId().getEndpoint(), request,  5000);
            if(resp == null) {
                log.error("{} set WorkerState is error", resp);
            }
            // log.debug("follow is {}", resp);
            return;
        } catch (InterruptedException | RemotingException e) {
            e.printStackTrace();
        }
    }
 
    /**
     * 同步更新整个State
     * @param s
     * @throws InterruptedException
     * @throws RemotingException
     */
    public void updateFsmState(State s) throws InterruptedException, RemotingException {
        if(isLeader()) {
            var closure =  new SyncClosure<State>() {
                @Override
                public void run(Status status) {
                    // log.debug("leader update {}", status);
                    synclock.notify();
                }
            };
            StateServerFactory.getStateServer().applyState(s, closure); 
            closure.synclock.wait(5000);
            return;
        }
 
        var ss = StateServerFactory.getStateServer();
        var request = new RequestState();
        request.setState(s);
        log.debug("my: {} leader id {}",ss.getCluster().getServerId(), ss.getNode().getLeaderId());
        ResponseSM resp;
        resp = (ResponseSM)StateServerFactory.getStateServer().getRpcClient().invokeSync(ss.getNode().getLeaderId().getEndpoint(), request,  5000);
        if(resp == null) {
            log.error("{} set State is error", resp);
        }
        // log.debug("follow is {}", resp);
        return;
      
    }


    public void updateFsmStateAsync(State s, Function<Status, Void> onCompleted) throws InterruptedException, RemotingException {
        if(isLeader()) {
            var closure =  new SyncClosure<State>() {
                @Override
                public void run(Status status) {
                    // log.debug("leader update {}", status);
                    if(status.isOk()) {
                        onCompleted.apply(status);
                    }
                }
            };
            StateServerFactory.getStateServer().applyState(s, closure); 
            return;
        }
 
        var ss = StateServerFactory.getStateServer();
        var request = new RequestState();
        request.setState(s);
        log.debug("my: {} leader id {}",ss.getCluster().getServerId(), ss.getNode().getLeaderId());
    
        StateServerFactory
        .getStateServer()
        .getRpcClient()
        .invokeAsync(ss.getNode().getLeaderId().getEndpoint(), request, new InvokeCallback() {

            @Override
            public void complete(Object result, Throwable err) {
                if(result != null){
                    onCompleted.apply(Status.OK());
                } else {
                    var status = new Status(10000, "rpc invokeAsync with request: {}", request);
                    log.debug("{}", status);
                    onCompleted.apply(status);
                }
            }
            
        }, 5000);
 
        // log.debug("follow is {}", resp);
        return;
      
    }


    public void applyState(State state, SyncClosure<State> closure) {
        // 所有的提交都必须再leader进行
        if (!ss.isLeader()) {
            ss.handlerNotLeaderError(closure);
            return;
        }
        
        try {
            closure.setValue(state);
            final Task task = new Task();
            task.setData(ByteBuffer.wrap(SerializerManager.getSerializer(SerializerManager.Hessian2).serialize(state)));
            task.setDone(closure); // 确认所有数据 一致, 不需要加锁
            StateServerFactory.getStateServer().getNode().apply(task);
        } catch (CodecException e) {
            String errorMsg = "Fail to encode TaskState";
            log.debug(errorMsg, e);
            closure.failure(errorMsg, PeerId.emptyPeer());
            closure.run(new Status(RaftError.EINTERNAL, errorMsg));
        }
    }

    public void applyWorkerState(WorkerState state, SyncClosure<State>  closure) {
        if (!ss.isLeader()) {
            ss.handlerNotLeaderError(closure);
            return;
        }

        
        useFsmStateAsync((fsmState)->{

            var wmap = fsmState.getWorkers();
            var wstate = wmap.get(state.getPeerId());
    
            if(wstate == null) {
                wmap.put(state.getPeerId(), state);
                try {
                    final Task task = new Task();
                    closure.setValue(fsmState);
                    task.setData(ByteBuffer.wrap(SerializerManager.getSerializer(SerializerManager.Hessian2).serialize(fsmState)));
                    task.setDone(closure);
                    StateServerFactory.getStateServer().getNode().apply(task); // 提交数据
                } catch (CodecException e) {
                    String errorMsg = "Fail to encode TaskState";
                    log.debug("{}:{}",errorMsg, e);
                    closure.failure(errorMsg, PeerId.emptyPeer());
                    closure.run(new Status(RaftError.EINTERNAL, errorMsg));
                }
            } else {
                closure.success(fsmState);
                closure.run(Status.OK());
            }
            return null;
        });
    

    
    
    }

    public ResponseSM redirect() {
        final ResponseSM response = new ResponseSM();
        response.setSuccess(false);
        if (this.node != null) {
            final PeerId leader = this.node.getLeaderId();
            if (leader != null) {
                response.setRedirect(leader);
            }
        }
        return response;
    }

    public void handlerNotLeaderError(final SyncClosure closure) {
        closure.failure("Not leader.", redirect().getRedirect());
        closure.run(new Status(RaftError.EPERM, "Not leader"));
    }


    private Executor createReadIndexExecutor() {
        //        final StoreEngineOptions opts = new StoreEngineOptions();
        //        return StoreEngineHelper.createReadIndexExecutor(opts.getReadIndexCoreThreads());
        
                return ThreadPoolUtil.newBuilder() //
                        .poolName("ReadIndexPool") //
                        .enableMetric(true) //
                        .coreThreads(4) //
                        .maximumThreads(4) //
                        .keepAliveSeconds(60L) //
                        .workQueue(new SynchronousQueue<>()) //
                        .threadFactory(new NamedThreadFactory("ReadIndexService", true)) //
                        .rejectedHandler(new ThreadPoolExecutor.AbortPolicy()) //
                        .build();
            }
}



    public static void main(String[] args) throws InterruptedException, RemotingException {
        var rpcClient = new BoltRaftRpcFactory().createRpcClient();
        rpcClient.init(new CliOptions());
        var resp = rpcClient.invokeSync(new Endpoint("localhost",4441), new RequestState(), 5000);
        log.info("{}", resp);
    }
}
